from .fileio import (
    copyfile,
    creation_date,
    get_file_hash,
    get_tmpdir,
    get_tmpfile,
    mkdir_p,
    mkdirp,
    print_json,
    read_file,
    read_json,
    read_yaml,
    recursive_find,
    workdir,
    write_file,
    write_json,
    write_yaml,
)
from .misc import chunks, get_hash, mb_to_bytes, print_bytes, slugify
from .terminal import confirm_action, get_installdir



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

# Default templates retrieved from:
# https://docs.aws.amazon.com/eks/latest/userguide/creating-a-vpc.html
vpc_template = "https://s3.us-west-2.amazonaws.com/amazon-eks/cloudformation/2020-10-29/amazon-eks-ipv6-vpc-public-private-subnets.yaml"

# EKS node group (workers template)
# https://docs.aws.amazon.com/eks/latest/userguide/retrieve-ami-id.html
workers_template = "https://s3.us-west-2.amazonaws.com/amazon-eks/cloudformation/2022-12-23/amazon-eks-nodegroup.yaml"

# Other useful links for AMI
# https://docs.aws.amazon.com/eks/latest/userguide/retrieve-ami-id.html
# https://docs.aws.amazon.com/eks/latest/userguide/eks-optimized-amis.html
# aws ssm get-parameter --name /aws/service/eks/optimized-ami/1.26/amazon-linux-2/recommended/image_id --region us-east-1 --query "Parameter.Value" --output text

# Can't use jinja2 here, we actually need that double {{}}
auth_config_data = """apiVersion: v1
kind: ConfigMap
metadata:
  name: aws-auth
  namespace: kube-system
data:
  mapRoles: |
    - rolearn: %s
      username: system:node:{{EC2PrivateDNSName}}
      groups:
        - system:bootstrappers
        - system:nodes
"""



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import copy


def chunks(listing, chunk_size):
    """
    Yield successive chunks from listing. Chunkify!
    """
    for i in range(0, len(listing), chunk_size):
        yield listing[i : i + chunk_size]


def slugify(name):
    """
    Slugify a name, replacing spaces with - and lowercase.
    """
    for char in [" ", ":", "/", "\\"]:
        name = name.replace(char, "-")
    return name.lower()


def print_bytes(byt, suffix="B"):
    """
    Pretty format size in bytes
    """
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(byt) < 1024.0:
            return f"{byt:3.1f} {unit}{suffix}"
        byt /= 1024.0
    return f"{byt:.1f} Yi{suffix}"


def mb_to_bytes(mb):
    """
    Convert mb to bytes, usually so we can derive a better format.
    """
    return mb * (1048576)


def get_hash(obj):
    """
    Get a hash for a random object (set, tuple, list, dict)

    All nested attributes must at least be hashable!
    """
    if isinstance(obj, (set, tuple, list)):
        return tuple([get_hash(o) for o in obj])
    if not isinstance(obj, dict):
        return hash(obj)
    copied = copy.deepcopy(obj)
    for k, v in copied.items():
        copied[k] = get_hash(v)
    return hash(tuple(frozenset(sorted(copied.items()))))



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import os

import kubescaler.utils as utils

install_dir = utils.get_installdir()
reps = {"$install_dir": install_dir, "$root_dir": os.path.dirname(install_dir)}

# User home
userhome = os.path.expanduser("~/.kubescaler")

# Default Kubernetes version (aws doesn't have 1.27)
kubernetes_version = 1.27

# The default GitHub registry with recipes (for docgen)
github_url = "https://github.com/converged-computing/kubescaler"



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import os


def get_installdir():
    """get_installdir returns the installation directory of the application"""
    return os.path.abspath(os.path.dirname(os.path.dirname(__file__)))


def confirm_action(question, force=False):
    """
    confirm if the user wants to perform a certain action

    Parameters
    ==========
    question: the question that will be asked
    force: if the user wants to skip the prompt
    """
    if force is True:
        return True

    response = input(question + " (yes/no)? ")
    while len(response) < 1 or response[0].lower().strip() not in "ynyesno":
        response = input("Please answer yes or no: ")

    if response[0].lower().strip() in "no":
        return False

    return True



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import os

import kubescaler.defaults as defaults
from kubescaler.utils import write_json


class Cluster:
    """
    A base cluster controller for scaling.
    """

    def __init__(
        self,
        name=None,
        description=None,
        tags=None,
        region=None,
        node_count=2,
        sleep_seconds=3,
        sleep_multiplier=1,
        max_nodes=3,
        min_nodes=0,
        machine_type=None,
        kubernetes_version=None,
    ):
        """
        A simple class to control creating a cluster
        """
        self.node_count = node_count

        # List or dict depending on cloud
        self.tags = tags
        self.name = os.path.basename(name or "kubescaler-cluster")
        self.max_nodes = max_nodes
        self.min_nodes = max(0, min_nodes)
        self.description = description or "A Kubescaler testing cluster"
        self.sleep_seconds = sleep_seconds
        self.kubernetes_version = kubernetes_version or defaults.kubernetes_version
        self.machine_type = machine_type

        # Sleep time multiplication factor must be > 1, defaults to 1.5
        self.sleep_multiplier = max(sleep_multiplier or 1, 1)
        self.sleep_time = sleep_seconds or 2

        # Region or default region
        self.region = region or self.default_region

        # Easy way to save times
        self.times = {}

    def delete_cluster(self):
        """
        Delete the cluster
        """
        raise NotImplementedError

    def save(self, results_file):
        """
        Save results to file.
        """
        write_json(self.data, results_file)

    @property
    def data(self):
        """
        Combine class data into json object to save
        """
        raise NotImplementedError

    def scale(self, *args, **kwargs):
        """
        Make a request to scale the cluster
        """
        raise NotImplementedError

    def resize_cluster(self, *args, **kwargs):
        """
        Do the resize of the cluster
        """
        raise NotImplementedError

    def create_cluster(self):
        """
        Create a cluster, with hard coded variables for now.
        """
        raise NotImplementedError



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

__version__ = "0.0.11"
AUTHOR = "Vanessa Sochat"
EMAIL = "vsoch@users.noreply.github.com"
NAME = "kubescaler"
PACKAGE_URL = "https://github.com/converged-computing/kubescaler"
KEYWORDS = "Kubernetes, elasticity, scaling, EKS, GKE"
DESCRIPTION = "Helper classes for scaling Kubernetes clusters"
LICENSE = "LICENSE"

################################################################################
# Global requirements

# Since we assume wanting Singularity and lmod, we require spython and Jinja2

INSTALL_REQUIRES = (
    ("ruamel.yaml", {"min_version": None}),
    ("jsonschema", {"min_version": None}),
    ("kubernetes", {"min_version": None}),
)

AWS_REQUIRES = (
    ("awscli", {"min_version": None}),
    ("boto3", {"min_version": None}),
)

# Prefer discovery clients - more control
GOOGLE_CLOUD_REQUIRES = (
    ("google-cloud-container", {"min_version": None}),
    ("google-api-python-client", {"min_version": None}),
)

TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)

################################################################################
# Submodule Requirements (versions that include database)

INSTALL_REQUIRES_ALL = (
    INSTALL_REQUIRES + GOOGLE_CLOUD_REQUIRES + AWS_REQUIRES + TESTS_REQUIRES
)



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import json
import os
import subprocess
import sys
import time

try:
    import boto3
except ImportError:
    sys.exit("Please pip install kubescaler[aws]")

from kubernetes import client as k8s
from kubernetes import config
from kubernetes import utils as k8sutils

import kubescaler.utils as utils
from kubescaler.cluster import Cluster
from kubescaler.decorators import retry, timed
from kubescaler.logger import logger

from .ami import get_latest_ami
from .template import auth_config_data, vpc_template, workers_template

stack_failure_options = ["DELETE", "DO_NOTHING", "ROLLBACK"]


class EKSCluster(Cluster):
    """
    A scaler for an Amazon EKS Cluster
    """

    default_region = "us-east-1"

    def __init__(
        self,
        name,
        admin_role_name=None,
        kube_config_file=None,
        keypair_name=None,
        keypair_file=None,
        on_stack_failure="DELETE",
        stack_timeout_minutes=15,
        auth_config_file=None,
        **kwargs,
    ):
        """
        Create an Amazon Cluster
        """
        super().__init__(name=name, **kwargs)

        # name for K8s IAM role
        self.admin_role_name = admin_role_name or "EKSServiceAdmin"

        # Secrets files
        self.keypair_name = keypair_name or "workers-pem"
        self.keypair_file = keypair_file or "aws-worker-secret.pem"
        self.auth_config_file = auth_config_file or "aws-auth-config.yaml"

        # You might want to update this to better debug (so not deleted)
        # DO_NOTHING | ROLLBACK | DELETE
        self.set_stack_failure(on_stack_failure)
        self.stack_timeout_minutes = max(1, stack_timeout_minutes)

        # Here we define cluster name from name
        self.cluster_name = self.name
        self.tags = self.tags or {}
        if not isinstance(self.tags, dict):
            raise ValueError("Tags must be key value pairs (dict)")

        # kube config file
        self.kube_config_file = kube_config_file or "kubeconfig-aws.yaml"
        self.image_ami = get_latest_ami(self.region, self.kubernetes_version)
        self.machine_type = self.machine_type or "m5.large"

        # Client connections
        self.session = boto3.Session(region_name=self.region)
        self.ec2 = self.session.client("ec2")
        self.cf = self.session.client("cloudformation")
        self.iam = self.session.client("iam")
        self.eks = self.session.client("eks")

        # Will be set later!
        self.workers_stack = None
        self.vpc_stack = None
        self.vpc_security_group = None
        self.vpc_subnet_private = None
        self.vpc_subnet_public = None
        self.vpc_id = None
        self.set_roles()

    def set_stack_failure(self, on_stack_failure):
        """
        Set the action to take if a stack fails to create.
        """
        self.on_stack_failure = on_stack_failure
        if self.on_stack_failure not in stack_failure_options:
            options = " | ".join(stack_failure_options)
            raise ValueError(
                f"{on_stack_failure} is not a valid option, choices are: {options}"
            )

    @timed
    def create_cluster(self):
        """
        Create a cluster
        """
        self.set_vpc_stack()
        self.set_subnets()

        try:
            cluster = self.eks.describe_cluster(name=self.cluster_name)
        except Exception:
            cluster = self.new_cluster()

        # Get the status and confirm it's active
        status = cluster["cluster"]["status"]
        if status != "ACTIVE":
            raise ValueError(
                f"Found cluster {self.cluster_name} but status is {status} and should be ACTIVE"
            )

        # Get cluster endpoint and security info so we can make kubectl config
        self.certificate = cluster["cluster"]["certificateAuthority"]["data"]
        self.endpoint = cluster["cluster"]["endpoint"]

        # Ensure we have a config to interact with, and write the keypair file
        self.ensure_kube_config()
        self.get_keypair()

        # The cluster is actually created with no nodes - just the control plane!
        # Here is where we create the workers, via a stack. Because apparently
        # AWS really likes their pancakes.
        self.set_workers_stack()
        self.create_auth_config()

        # We can only wait for the node group after we set the auth config!
        # I was surprised this is expecting the workers name and not the node
        # group name.
        self.wait_for_nodes()

        print(f"🦊️ Writing config file to {self.kube_config_file}")
        print(f"  Usage: kubectl --kubeconfig={self.kube_config_file} get nodes")
        return cluster

    @timed
    def wait_for_nodes(self):
        """
        Wait for the nodes to be ready.

        We do this separately to allow timing. This function would be improved if
        we didn't need subprocess, but the waiter doesn't seem to work.
        """
        while True:
            print(f"Waiting for {self.node_count} nodes to be Ready...")
            time.sleep(5)
            # TODO fix to not require doing this! Maybe someone else wants to take a look?
            res = subprocess.check_output(
                [
                    "kubectl",
                    f"--kubeconfig={self.kube_config_file}",
                    "--no-headers=true",
                    "get",
                    "nodes",
                ]
            )
            lines = [x.strip() for x in res.decode("utf-8").split("\n") if x.strip()]
            ready_count = 0
            for line in lines:
                if "NotReady" in line:
                    continue
                ready_count += 1
            if ready_count == self.node_count:
                break

        # The waiter doesn't seem to work - so we call kubectl until it's ready
        # waiter = self.eks.get_waiter("nodegroup_active")
        # waiter.wait(clusterName=self.cluster_name, nodegroupName=self.node_autoscaling_group_name)

    def create_auth_config(self):
        """
        Deploy a config map that tells the master how to contact the workers

        After this, kubectl --kubeconfig=./kubeconfig.yaml get nodes
        will (or I should say "should") work!
        """
        # Easier to write to file and then apply!
        auth_config = auth_config_data % self.node_instance_role
        utils.write_file(auth_config, self.auth_config_file)

        # We can likely do it this way in the future (will open issue)
        config.load_kube_config(self.kube_config_file)
        koobcli = k8s.ApiClient()

        # If the cluster was already created, this would raise an error
        try:
            k8sutils.create_from_yaml(koobcli, self.auth_config_file)
        except Exception:
            pass

    def ensure_kube_config(self):
        """
        Ensure the kubernetes kubectl config file exists

        Since this might change, let's always just write it again.
        We require the user to install awscli so the aws executable
        should be available.
        """
        cluster_config = {
            "apiVersion": "v1",
            "kind": "Config",
            "clusters": [
                {
                    "cluster": {
                        "server": str(self.endpoint),
                        "certificate-authority-data": str(self.certificate),
                    },
                    "name": "kubernetes",
                }
            ],
            "contexts": [
                {"context": {"cluster": "kubernetes", "user": "aws"}, "name": "aws"}
            ],
            "current-context": "aws",
            "preferences": {},
            "users": [
                {
                    "name": "aws",
                    "user": {
                        "exec": {
                            "apiVersion": "client.authentication.k8s.io/v1beta1",
                            "command": "aws",
                            "args": [
                                "--region",
                                self.region,
                                "eks",
                                "get-token",
                                "--cluster-name",
                                self.cluster_name,
                            ],
                        }
                    },
                }
            ],
        }
        utils.write_yaml(cluster_config, self.kube_config_file)

    def get_keypair(self):
        """
        Write keypair file.
        """
        try:
            # Check if keypair exists, if not, ignore this step.
            return self.ec2.describe_key_pairs(KeyNames=[self.keypair_name])
        except Exception:
            return self.create_keypair()

    def create_keypair(self):
        """
        Create the keypair secret and associated file.
        """
        key = self.ec2.create_key_pair(KeyName=self.keypair_name)
        private_key = key["KeyMaterial"]

        # Write to file - this needs to be managed by client runner
        # to ensure uniqueness of names (and not rewriting existing files)
        utils.write_file(private_key, self.keypair_file)
        os.chmod(self.keypair_file, 400)
        return key

    def set_workers_stack(self):
        """
        Get or create the workers stack, or the nodes for the cluster.
        """
        try:
            self.workers_stack = self.cf.describe_stacks(StackName=self.workers_name)
        except Exception:
            self.workers_stack = self.create_workers_stack()

        # We need this role to later associate master with workers
        self.node_instance_role = None
        for output in self.workers_stack["Stacks"][0]["Outputs"]:
            if output["OutputKey"] == "NodeInstanceRole":
                self.node_instance_role = output["OutputValue"]
            if output["OutputKey"] == "NodeAutoScalingGroup":
                self.node_autoscaling_group_name = output["OutputValue"]

    @timed
    def delete_workers_stack(self):
        """
        Delete the workers stack.
        """
        return self.delete_stack(self.workers_name)

    @timed
    def delete_vpc_stack(self):
        """
        Delete the vpc stack
        """
        return self.delete_stack(self.vpc_name)

    @timed
    def create_workers_stack(self):
        """
        Create the workers stack (the nodes for the EKS cluster)
        """
        stack = self.cf.create_stack(
            StackName=self.workers_name,
            TemplateURL=workers_template,
            Capabilities=["CAPABILITY_IAM"],
            Parameters=[
                {"ParameterKey": "ClusterName", "ParameterValue": self.cluster_name},
                {
                    "ParameterKey": "ClusterControlPlaneSecurityGroup",
                    "ParameterValue": self.vpc_security_group,
                },
                {
                    "ParameterKey": "NodeGroupName",
                    "ParameterValue": self.node_group_name,
                },
                {
                    "ParameterKey": "NodeAutoScalingGroupMinSize",
                    "ParameterValue": str(self.min_nodes),
                },
                {
                    "ParameterKey": "NodeAutoScalingGroupDesiredCapacity",
                    "ParameterValue": str(self.node_count),
                },
                {
                    "ParameterKey": "NodeAutoScalingGroupMaxSize",
                    "ParameterValue": str(self.max_nodes),
                },
                {
                    "ParameterKey": "NodeInstanceType",
                    "ParameterValue": self.machine_type,
                },
                {"ParameterKey": "NodeImageId", "ParameterValue": self.image_ami},
                {"ParameterKey": "KeyName", "ParameterValue": self.keypair_name},
                {"ParameterKey": "VpcId", "ParameterValue": self.vpc_id},
                {
                    "ParameterKey": "Subnets",
                    "ParameterValue": ",".join(self.vpc_subnet_ids),
                },
            ],
            TimeoutInMinutes=self.stack_timeout_minutes,
            OnFailure=self.on_stack_failure,
        )
        print(self.stack_timeout_minutes)
        return self._create_stack(stack, self.workers_name)

    def new_cluster(self):
        """
        Create a new cluster.
        """
        # Create Kubernetes cluster.
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/eks/client/create_cluster.html
        self.eks.create_cluster(
            name=self.cluster_name,
            version=str(self.kubernetes_version),
            roleArn=self.role_arn,
            tags=self.tags,
            resourcesVpcConfig={
                "subnetIds": self.vpc_subnet_ids,
                "securityGroupIds": [self.vpc_security_group],
            },
        )
        logger.info("⭐️ Cluster creation started! Waiting...")
        waiter = self.eks.get_waiter("cluster_active")
        waiter.wait(name=self.cluster_name)

        # When it's ready, save the cluster
        return self.eks.describe_cluster(name=self.cluster_name)

    def set_vpc_stack(self):
        """
        Get the stack
        """
        # Does it already exist?
        try:
            self.vpc_stack = self.cf.describe_stacks(StackName=self.vpc_name)
        except Exception:
            self.vpc_stack = self.create_vpc_stack()

    def create_vpc_stack(self):
        """
        Create a new stack from the template
        """
        # If not, create it from the template
        stack = self.cf.create_stack(
            StackName=self.vpc_name,
            TemplateURL=vpc_template,
            Parameters=[],
            TimeoutInMinutes=self.stack_timeout_minutes,
            OnFailure=self.on_stack_failure,
        )
        return self._create_stack(stack, self.vpc_name)

    def _create_stack(self, stack, stack_name):
        """
        Shared function to check validity of stack and wait!

        I didn't add retry here, because it usually fails for some
        "good" reason (e.g., not enough network)
        """
        if stack is None:
            raise ValueError("Could not create stack")

        if "StackId" not in stack:
            raise ValueError("Could not create VPC stack")

        try:
            logger.info(f"Waiting for {stack_name} stack...")
            waiter = self.cf.get_waiter("stack_create_complete")
            waiter.wait(StackName=stack_name)
        except Exception:
            raise ValueError("Waiting for stack creation exceeded wait time.")

        # Retrieve the same metadata if we had retrieved it
        return self.cf.describe_stacks(StackName=stack_name)

    def delete_stack(self, stack_name):
        """
        Delete a stack and wait for it to be deleted
        """
        logger.info(f"🥞️ Attempting delete of stack {stack_name}...")
        try:
            self.cf.delete_stack(StackName=stack_name)
        except Exception:
            logger.warning(f"Stack {stack_name} does not exist.")
            return
        try:
            logger.info(f"Waiting for {stack_name} to be deleted..")
            waiter = self.cf.get_waiter("stack_delete_complete")
            waiter.wait(StackName=stack_name)
        except Exception:
            raise ValueError("Waiting for stack deletion exceeded wait time.")

    def set_roles(self):
        """
        Create the default IAM arn role for the admin
        """
        try:
            # See if role exists.
            self.role = self.iam.get_role(RoleName=self.admin_role_name)
        except Exception:
            self.role = self.create_role()
        self.role_arn = self.role["Role"]["Arn"]

    def create_role(self):
        """
        Create the role for eks
        """
        # This is an AWS role policy document.  Allows access for EKS.
        policy_doc = json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Action": "sts:AssumeRole",
                        "Effect": "Allow",
                        "Principal": {"Service": "eks.amazonaws.com"},
                    }
                ],
            }
        )

        # Create role and attach needed policies for EKS
        role = self.iam.create_role(
            RoleName=self.admin_role_name,
            AssumeRolePolicyDocument=policy_doc,
            Description="Role providing access to EKS resources from EKS",
        )

        self.iam.attach_role_policy(
            RoleName=self.admin_role_name,
            PolicyArn="arn:aws:iam::aws:policy/AmazonEKSClusterPolicy",
        )

        self.iam.attach_role_policy(
            RoleName=self.admin_role_name,
            PolicyArn="arn:aws:iam::aws:policy/AmazonEKSServicePolicy",
        )
        return role

    def set_subnets(self):
        """
        Create VPC subnets
        """
        if not self.vpc_stack:
            raise ValueError("set_subnets needs to be called after stack creation.")

        # Unwrap list of outputs into values we care about.
        for output in self.vpc_stack["Stacks"][0]["Outputs"]:
            if output["OutputKey"] == "SecurityGroups":
                self.vpc_security_group = output["OutputValue"]
            if output["OutputKey"] == "VPC":
                self.vpc_id = output["OutputValue"]
            if output["OutputKey"] == "SubnetsPublic":
                self.vpc_subnet_public = output["OutputValue"].split(",")
            if output["OutputKey"] == "SubnetsPrivate":
                self.vpc_subnet_private = output["OutputValue"].split(",")

    @property
    def vpc_subnet_ids(self):
        """
        Get listing of private and public subnet ids
        """
        vpc_subnet_ids = []
        if self.vpc_subnet_private is not None:
            vpc_subnet_ids += self.vpc_subnet_private
        if self.vpc_subnet_public is not None:
            vpc_subnet_ids += self.vpc_subnet_public
        return vpc_subnet_ids

    @property
    def vpc_name(self):
        return self.name + "-vpc"

    @property
    def workers_name(self):
        return self.name + "-workers"

    @property
    def node_group_name(self):
        return self.cluster_name + "-worker-group"

    @timed
    def delete_cluster(self):
        """
        Delete the cluster

        Let's be conservative and leave the kube config files, because if
        something goes wrong we want to be able to interact with them.
        And let's go backwards - deleting first what we created last.
        """
        logger.info("🔨️ Deleting node workers...")
        self.delete_workers_stack()
        # We could delete keypair, but let's keep for now, assuming could be reused elsewhere
        # and a deletion might be unexpected to the user

        # Now delete the cluster
        try:
            self.eks.delete_cluster(name=self.cluster_name)
        except Exception as e:
            logger.info(f"⏳️ Cluster likely already deleted: {e}")
            return

        logger.info("⏳️ Cluster deletion started! Waiting...")
        waiter = self.eks.get_waiter("cluster_deleted")
        waiter.wait(name=self.cluster_name)

        # Delete the VPC stack and we are done!
        logger.info("🥅️ Deleting VPC and associated assets...")
        self.delete_vpc_stack()
        logger.info("⭐️ Done!")

    @property
    def data(self):
        """
        Combine class data into json object to save
        """
        return {
            "times": self.times,
            "cluster_name": self.cluster_name,
            "machine_type": self.machine_type,
            "name": self.name,
            "region": self.region,
            "tags": self.tags,
            "description": self.description,
        }

    @retry
    def scale(self, count):
        """
        Make a request to scale the cluster
        """
        response = self.cf.update_stack(
            StackName=self.workers_name,
            UsePreviousTemplate=True,
            Capabilities=["CAPABILITY_IAM"],
            Parameters=[
                {
                    "ParameterKey": "NodeAutoScalingGroupDesiredCapacity",
                    "ParameterValue": str(count),
                },
                {"ParameterKey": "ClusterName", "UsePreviousValue": True},
                {
                    "ParameterKey": "ClusterControlPlaneSecurityGroup",
                    "ParameterValue": self.vpc_security_group,
                },
                {
                    "ParameterKey": "NodeGroupName",
                    "ParameterValue": self.node_group_name,
                },
                {
                    "ParameterKey": "NodeAutoScalingGroupMinSize",
                    "ParameterValue": str(self.min_nodes),
                },
                {
                    "ParameterKey": "NodeAutoScalingGroupDesiredCapacity",
                    "ParameterValue": str(count),
                },
                {
                    "ParameterKey": "NodeAutoScalingGroupMaxSize",
                    "ParameterValue": str(self.max_nodes),
                },
                {"ParameterKey": "KeyName", "ParameterValue": self.keypair_name},
                {"ParameterKey": "VpcId", "ParameterValue": self.vpc_id},
                {
                    "ParameterKey": "Subnets",
                    "ParameterValue": ",".join(self.vpc_subnet_ids),
                },
            ],
        )

        # Wait for stack update to be complete. Note this does not seem
        # to work. Instead we update the node count and then wait for the nodes.
        # waiter = self.cf.get_waiter('stack_update_complete')
        # waiter.wait(StackName=self.workers_name)

        # If successful, save new node count
        self.node_count = count
        self.wait_for_nodes()
        return response



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import base64
import sys
import tempfile
import time

from kubernetes import client as kubernetes_client

from kubescaler.cluster import Cluster
from kubescaler.decorators import retry, timed

try:
    import google.auth
    import google.auth.transport.requests
    from google.api_core.exceptions import NotFound
    from google.cloud import container_v1
except ImportError:
    sys.exit("Please pip install kubescaler[google]")


class GKECluster(Cluster):
    """
    A scaler for a Google Kubernetes Engine (GKE) cluster
    """

    default_region = "us-central1-a"

    def __init__(
        self,
        project,
        machine_type_memory_gb=32,
        machine_type_vcpu=8,
        **kwargs,
    ):
        """
        A simple class to control creating a cluster
        """
        super().__init__(**kwargs)

        # This client we can use to interact with Google Cloud GKE
        # https://github.com/googleapis/python-container/blob/main/google/cloud/container_v1/services/cluster_manager/client.py#L96
        print("⭐️ Creating global cluster manager client...")
        self.client = container_v1.ClusterManagerClient()
        self.project = project
        self.machine_type = self.machine_type or "c2-standard-8"
        self.machine_type_vcpu = machine_type_vcpu
        self.machine_type_memory_gb = machine_type_memory_gb
        self.tags = self.tags or ["kubescaler-cluster"]
        self.configuration = None

    @timed
    def delete_cluster(self):
        """
        Delete the cluster
        """
        request = container_v1.DeleteClusterRequest(name=self.cluster_name)
        # Make the request, and check until deleted!
        self.client.delete_cluster(request=request)
        self.configuration = None
        self.wait_for_delete()

    @property
    def zone(self):
        """
        The region is the zone minus the last letter!
        """
        return self.region.rsplit("-", 1)[0]

    @property
    def data(self):
        """
        Combine class data into json object to save
        """
        return {
            "times": self.times,
            "cluster_name": self.cluster_name,
            "name": self.name,
            "machine_type": self.machine_type,
            "region": self.region,
            "tags": self.tags,
            "description": self.description,
        }

    def scale_up(self, count, pool_name="default-pool"):
        """
        Make a request to scale the cluster
        """
        return self.scale(count, count, count + 1, pool_name=pool_name)

    def scale_down(self, count, pool_name="default-pool"):
        """
        Make a request to scale the cluster
        """
        return self.scale(
            count, max(count - 1, self.min_nodes), count, pool_name=pool_name
        )

    def scale(self, count, min_count, max_count, pool_name="default-pool"):
        """
        Make a request to scale the cluster
        """
        node_pool_name = f"{self.cluster_name}/nodePools/{pool_name}"

        # Always make the max node count one more than we want
        # I'm not sure if we need to change the policy with the size
        autoscaling = container_v1.NodePoolAutoscaling(
            enabled=True,
            min_node_count=min_count,
            max_node_count=max_count,
            #            total_min_node_count=count,
            #            total_max_node_count=count,
        )

        # https://github.com/googleapis/python-container/blob/main/google/cloud/container_v1/types/cluster_service.py#L3884
        request = container_v1.SetNodePoolAutoscalingRequest(
            autoscaling=autoscaling,
            name=node_pool_name,
        )
        self.client.set_node_pool_autoscaling(request=request)

        # This is wrapped in a retry
        self.resize_cluster(count, node_pool_name)

        # wait for it to be running again, will go from reconciling -> running
        return self.wait_for_status(2)

    @retry
    def resize_cluster(self, count, node_pool_name):
        """
        Do the resize of the cluster
        """
        # This is the request that actually changes the size
        request = container_v1.SetNodePoolSizeRequest(
            node_count=count,
            name=node_pool_name,
        )
        return self.client.set_node_pool_size(request=request)

    @property
    def node_config(self):
        """
        Create the node config

        Note that instead of initial_node_count + node_config above,
        we could just use node_pool. I think the first creates the second,
        and I'm not sure about pros/cons.
        """
        # Note that if you use GKE Autopilot you need to use a different class, see the link:
        # https://github.com/googleapis/python-container/blob/main/google/cloud/container_v1/types/cluster_service.py#L448
        node_config = container_v1.NodeConfig(
            machine_type=self.machine_type,
            tags=self.tags
            # metadata = {"startup-script": my_startup_script,
            #            "user-data": my_user_data}
        )
        print("\n🥣️ cluster node_config")
        print(node_config)
        return node_config

    def get_k8s_client(self):
        """
        Get a client to use to interact with the cluster, either corev1.api
        or the kubernetes api client.

        https://github.com/googleapis/python-container/issues/6
        """
        request = {"name": self.cluster_name}
        response = self.client.get_cluster(request=request)
        creds, projects = google.auth.default(
            scopes=["https://www.googleapis.com/auth/cloud-platform"]
        )
        auth_req = google.auth.transport.requests.Request()
        creds.refresh(auth_req)

        # Save the configuration for advanced users to user later
        if not self.configuration:
            configuration = kubernetes_client.Configuration()
            configuration.host = f"https://{response.endpoint}"
            with tempfile.NamedTemporaryFile(delete=False) as ca_cert:
                ca_cert.write(
                    base64.b64decode(response.master_auth.cluster_ca_certificate)
                )
                configuration.ssl_ca_cert = ca_cert.name
            configuration.api_key_prefix["authorization"] = "Bearer"
            configuration.api_key["authorization"] = creds.token
            self.configuration = configuration

        # This has .api_client for just the api client
        return kubernetes_client.CoreV1Api(
            kubernetes_client.ApiClient(self.configuration)
        )

    @property
    def cluster(self):
        """
        Get the cluster proto with our defaults
        """
        # Design our initial cluster!
        # https://github.com/googleapis/python-container/blob/main/google/cloud/container_v1/types/cluster_service.py#LL2119C1-L2124C1

        # Command for comparison
        # TODO I don't see where dataplane is, or cloud dns, can add later!
        # gcloud container clusters create flux-cluster \
        #   --region=us-central1-a --project $GOOGLE_PROJECT \
        #   --machine-type c2d-standard-112 --num-nodes=32 \
        #   --cluster-dns=clouddns --cluster-dns-scope=cluster \
        #   --tags=flux-cluster  --enable-dataplane-v2 \
        #   --threads-per-core=1
        # Note: we can add node_config for customizing node pools further

        # TODO these look useful / interesting
        # autoscaling (google.cloud.container_v1.types.ClusterAutoscaling):
        #  Cluster-level autoscaling configuration.

        # Autoscaling - try optimizing
        # PROFILE_UNSPECIFIED = 0
        # OPTIMIZE_UTILIZATION = 1
        # BALANCED = 2
        autoscaling_profile = container_v1.ClusterAutoscaling.AutoscalingProfile(1)

        # These are hard coded for c2-standard-8
        # https://cloud.google.com/compute/docs/compute-optimized-machines
        resource_limits = [
            container_v1.ResourceLimit(
                resource_type="cpu",
                minimum=self.machine_type_vcpu,
                maximum=self.machine_type_vcpu * self.max_nodes,
            ),
            container_v1.ResourceLimit(
                resource_type="memory",
                minimum=self.machine_type_memory_gb,
                maximum=self.machine_type_memory_gb * self.max_nodes,
            ),
        ]
        cluster_autoscaling = container_v1.ClusterAutoscaling(
            enable_node_autoprovisioning=True,
            autoprovisioning_locations=[self.zone],
            autoscaling_profile=autoscaling_profile,
            resource_limits=resource_limits,
        )

        # vertical_pod_autoscaling (google.cloud.container_v1.types.VerticalPodAutoscaling):
        #  Cluster-level Vertical Pod Autoscaling
        #  configuration.
        cluster = container_v1.Cluster(
            name=self.name,
            description=self.description,
            initial_node_count=self.node_count,
            node_config=self.node_config,
            autoscaling=cluster_autoscaling,
        )
        print("\n🥣️ cluster spec")
        print(cluster)
        return cluster

    @timed
    def create_cluster(self):
        """
        Create a cluster, with hard coded variables for now.
        """
        # https://github.com/googleapis/python-container/blob/main/google/cloud/container_v1/types/cluster_service.py#L3527
        request = container_v1.CreateClusterRequest(
            parent=f"projects/{self.project}/locations/{self.region}",
            cluster=self.cluster,
        )

        print("\n🥣️ cluster creation request")
        print(request)

        # Make the request
        response = self.client.create_cluster(request=request)
        print(response)

        # Status 2 is running (1 is provisioning)
        print(f"⏱️   Waiting for {self.cluster_name} to be ready...")
        return self.wait_for_status(2)

    @property
    def cluster_name(self):
        return f"projects/{self.project}/locations/{self.region}/clusters/{self.name}"

    def wait_for_delete(self):
        """
        Wait until the cluster is running (status 2 I think?)
        """
        sleep = self.sleep_time
        while True:
            time.sleep(sleep)

            # https://github.com/googleapis/python-container/blob/main/google/cloud/container_v1/types/cluster_service.py#L3569
            request = container_v1.GetClusterRequest(name=self.cluster_name)
            # Make the request
            try:
                self.client.get_cluster(request=request)
            except NotFound:
                return

            # Some other issue
            except Exception:
                raise
            sleep = sleep * self.sleep_multiplier

    def wait_for_status(self, status=2):
        """
        Wait until the cluster is running (status 2 I think?)

        status codes:
        provisioning: 1
        running: 2
        reconciling: 3
        stopping: 4
        """
        sleep = self.sleep_time

        # https://github.com/googleapis/python-container/blob/main/google/cloud/container_v1/types/cluster_service.py#L3569
        request = container_v1.GetClusterRequest(name=self.cluster_name)
        response = None

        while not response or response.status.value != status:
            if response:
                print(
                    f"Cluster {self.cluster_name} does not have status {status}, found {response.status}. Sleeping {sleep}"
                )
            time.sleep(sleep)

            # Make the request
            response = self.client.get_cluster(request=request)
            sleep = sleep * self.sleep_multiplier

        # Get it once more before returning (has complete size, etc)
        return self.client.get_cluster(request=request)



# Copyright 2022-2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import time
from functools import partial, update_wrapper


class timed:
    """
    Time the runtime of a function, add to times
    """

    def __init__(self, func):
        update_wrapper(self, func)
        self.func = func

    def __get__(self, obj, objtype):
        return partial(self.__call__, obj)

    def __call__(self, cls, *args, **kwargs):
        name = self.func.__name__

        # If it's not create or delete cluster we need the node count
        if name not in ["create_cluster", "delete_cluster"]:
            name = f"{name}-size-{cls.node_count}"
        start = time.time()
        res = self.func(cls, *args, **kwargs)
        end = time.time()
        cls.times[name] = round(end - start, 3)
        return res


class retry:
    """
    Retry a function that is part of a class
    """

    def __init__(self, func, attempts=5, timeout=2):
        update_wrapper(self, func)
        self.func = func
        self.attempts = attempts
        self.timeout = timeout

    def __get__(self, obj, objtype):
        return partial(self.__call__, obj)

    def __call__(self, cls, *args, **kwargs):
        attempt = 0
        attempts = self.attempts
        timeout = self.timeout
        while attempt < attempts:
            try:
                return self.func(cls, *args, **kwargs)
            except Exception as e:
                sleep = timeout + 3**attempt
                print(f"Retrying in {sleep} seconds - error: {e}")
                time.sleep(sleep)
                attempt += 1
        return self.func(cls, *args, **kwargs)



# Copyright 2023 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

import boto3
from dateutil import parser

# aws ssm get-parameter --name /aws/service/eks/optimized-ami/1.26/amazon-linux-2/recommended/image_id --region us-east-1 --query "Parameter.Value" --output text


def get_latest_ami(region, kubernetes_version=1.26):
    """
    Get the latest AMI for a specific region.
    """
    client = boto3.client("ec2", region_name=region)
    filters = [({"Name": "name", "Values": [f"*-eks-node-{kubernetes_version}*"]})]
    image_listing = client.describe_images(Filters=filters)
    latest = filter_images(image_listing["Images"])
    return latest["ImageId"]


def filter_images(images):
    """
    Filter images down to latest
    """
    latest = None
    for image in images:
        if not latest:
            latest = image
            continue
        if parser.parse(image["CreationDate"]) > parser.parse(latest["CreationDate"]):
            latest = image
    return latest
